from dataclasses import dataclass
import re
import warnings

class dsiParser:
    """Parser to parse D-SI unit string into a tree
    """
    def __init__(self, latexDefaultWrapper='$$', latexDefaultPrefix='', latexDefaultSuffix=''):
        """
        Args:
            latexDefaultWrapper (str, optional): String to be added both in the beginning and the end of the LaTeX string. Defaults to '$$'.
            latexDefaultPrefix (str, optional): String to be added in the beginning of the LaTeX string, after the wrapper. Defaults to ''.
            latexDefaultSuffix (str, optional): String to be added in the end of the LaTeX string, before the wrapper. Defaults to ''.
        """
        self.latexDefaultWrapper = latexDefaultWrapper
        self.latexDefaultPrefix = latexDefaultPrefix
        self.latexDefaultSuffix = latexDefaultSuffix

    def parse(self, dsiString: str):
        """parses a D-SI string into a tree structure


        Args:
            dsiString (str): D-SI unit raw string

        Raises:
            RuntimeWarning: double backslashes in D-SI string
            RuntimeWarning: empty D-SI string

        Returns:
            dsiTree: dsiTree object containing the D-SI unit
        """
        warningMessages = []
        # Catch any double (triple...) \ before they annoy us
        while r'\\' in dsiString:
            warningMessages.append(_warn(f"Double backslash found in string, treating as one backslash: \"{dsiString}\"", RuntimeWarning))
            dsiString = dsiString.replace(r'\\', '\\')

        if dsiString == "":
            warningMessages.append(_warn("Given D-SI string is empty!", RuntimeWarning))
            return _dsiTree(dsiString, [], warningMessages, self.latexDefaultWrapper, self.latexDefaultPrefix, self.latexDefaultSuffix)

        tree = []
        (tree, fractionWarnings) = self._parseDsiFraction(dsiString)
        warningMessages += fractionWarnings
        for i, node in enumerate(tree):
            (tree[i], fractionlessWarnings) = self._parseFractionlessDsi(node)
            warningMessages += fractionlessWarnings
        return _dsiTree(dsiString, tree, warningMessages, self.latexDefaultWrapper, self.latexDefaultPrefix, self.latexDefaultSuffix)
    
    def _parseDsiFraction(self, dsiString: str):
        """parses D-SI fraction into list of fraction elements

        Args:
            dsiString (str): D-SI unit raw string

        Raises:
            RuntimeWarning: String must not contain more than one "per", 
                            as defined in the D-SI specs

        Returns:
            list: strings separated by the "per"
            list: warning messages of problems encountered while parsing
        """
        tree = []
        warningMessages = []
        tree = dsiString.split(r"\per")
        if len(tree) > 2:
            warningMessages.append(_warn(r"The dsi string contains more than one \per, does not "+
                        f"match specs! Given string: {dsiString}", 
                        RuntimeWarning))
        return (tree, warningMessages)

    def _parseFractionlessDsi(self, dsiString:str):
        """parses D-SI unit string without fractions

        Args:
            dsiString (str): D-SI unit raw string, not containing any fractions

        Raises:
            RuntimeWarning: if string does not meet the specs

        Returns:
            list: list of nodes
            list: warning messages of problems encountered while parsing
        """
        warningMessages = []
        items = dsiString.split("\\")
        if items[0] == '': #first item of List should be empty, remove it
            items.pop(0) 
        else:
            warningMessages.append(_warn(f"string should start with \\, string given was \"{dsiString}\"", RuntimeWarning))
        nodes = []

        (prefix, unit, exponent) = ('','','')
        item = items.pop(0)
        while True:
            if item in _dsiPrefixesLatex:
                prefix = item
                try:
                    item = items.pop(0)
                except IndexError:
                    item = ''
            if item in _dsiUnitsLatex:
                unit = item
                try:
                    item = items.pop(0)
                except IndexError:
                    item = ''
            if re.match(r'tothe\{-?\d+\.?\d?\}', item):
                exponent = item.split('{')[1].split('}')[0]
                try:
                    item = items.pop(0)
                except IndexError:
                    item = ''
            elif re.match(r'tothe\{.*\}', item):
                exponent = item.split('{')[1].split('}')[0]
                _warn(f"The exponent \"{exponent}\" is not a number!", RuntimeWarning)
                try:
                    item = items.pop(0)
                except IndexError:
                    item = ''
            if (prefix, unit, exponent) == ('','',''):
                unit = item
                try:
                    item = items.pop(0)
                except IndexError:
                    item = ''
                warningMessages.append(_warn(f"The identifier \"{unit}\" does not match any D-SI units!", RuntimeWarning))
            elif unit == '':
                itemStr = ""
                if prefix != "":
                    itemStr = itemStr + "\\" + prefix
                if exponent != "":
                    itemStr = itemStr + r"\tothe{" + exponent + r"}"
                warningMessages.append(_warn(f"This D-SI unit seems to be missing the base unit! \"{itemStr}\"", RuntimeWarning))

            nodes.append(_node(prefix, unit, exponent))
            if (len(items) == 0) and (item == ''): break
            (prefix, unit, exponent) = ('','','')
        
        return (nodes, warningMessages)

class _dsiTree:
    """D-SI representation in tree form, also includes validity check and warnings about D-SI string.
       Tree format: list of lists:
           List format: 
           First layer: items of the fraction 
           Second layer: nodes containing prefix, unit, power
    """
    def __init__(self, dsiString: str, dsiTree, warningMessages, latexDefaultWrapper='$$', latexDefaultPrefix='', latexDefaultSuffix=''):
        """
        Args:
            dsiString (str): the D-SI unit string to be parsed
        """
        self.dsiString = dsiString
        self.tree = dsiTree
        self.warnings = warningMessages
        self.valid = len(self.warnings) == 0
        self._latexDefaultWrapper = latexDefaultWrapper
        self._latexDefaultPrefix = latexDefaultPrefix
        self._latexDefaultSuffix = latexDefaultSuffix


    def toLatex(self, wrapper=None, prefix=None, suffix=None):
        """converts D-SI unit string to LaTeX

        Args:
            wrapper (str, optional): String to be added both in the beginning and the end of the LaTeX string. Defaults to the value set in the parser object.
            prefix (str, optional): String to be added in the beginning of the LaTeX string, after the wrapper. Defaults to the value set in the parser object.
            suffix (str, optional): String to be added in the end of the LaTeX string, before the wrapper. Defaults to the value set in the parser object.

        Returns:
            str: the corresponding LaTeX code
        """
        
        # If no wrapper/prefix/suffix was given, set to the parser's default
        wrapper = self._latexDefaultWrapper if wrapper == None else wrapper
        prefix = self._latexDefaultPrefix if prefix == None else prefix
        suffix = self._latexDefaultSuffix if suffix == None else suffix

        if self.tree == []:
            if len(prefix)+len(suffix) > 0:
                return wrapper+prefix+suffix+wrapper
            else:
                return ""
        latexArray = []
        if len(self.tree) == 1: # no fractions
            for node in self.tree[0]:
                latexArray.append(node.toLatex())
            latexString = r'\,'.join(latexArray)
        elif len(self.tree) == 2: # one fraction
            latexString = ""
            latexString += r'\frac'
            for frac in self.tree:
                latexString += r'{'
                nodeArray = []
                for node in frac:
                    nodeArray.append(node.toLatex())
                latexString += r'\,'.join(nodeArray)
                latexString += r'}'
        else: # more than one fraction 
            latexString = ""
            for i in range(len(self.tree)):
                nodeArray = []
                if i > 0:
                    latexString += r'{\color{red}/}'
                for node in self.tree[i]:
                    nodeArray.append(node.toLatex())
                latexString += r'\,'.join(nodeArray)
        return wrapper+prefix+latexString+suffix+wrapper
        
@dataclass
class _node:
    """one node of the D-SI tree, containing prefix, unit, power
    """
    prefix: str
    unit: str
    exponent: str

    def toLatex(self):
        """generates a latex string from a node

        Returns:
            str: latex representation
        """
        latexString = ""
        if self.prefix:
            latexString += _dsiPrefixesLatex[self.prefix]
        try:
            latexString += _dsiUnitsLatex[self.unit]
        except KeyError:
            latexString += r'{\color{red}\mathrm{'+self.unit+r'}}'
        if self.exponent:
            if self.exponent == '0.5':
                latexString = r'\sqrt{' + latexString + r'}'
            elif re.match(r'-?\d+\.?\d?', self.exponent):
                latexString += r'^{' + self.exponent + r'}'
            else:
                latexString += r'^{{\color{red}\mathrm{' + self.exponent + r'}}}'
        
        if self.unit == "":
            latexString = r'{\color{red}'+latexString+r'}'

        return latexString

def _warn(message: str, warningClass):
    """Output warning on command line and return warning message

    Args:
        message (str): warning message
        warningClass: Warning type

    Returns:
        str: message
    """
    warnings.warn(message, warningClass)
    return message
            
# mapping D-SI prefixes to latex
_dsiPrefixesLatex = {
    'deca': r'\mathrm{da}',
    'hecto': r'\mathrm{h}',
    'kilo': r'\mathrm{k}',
    'mega': r'\mathrm{M}',
    'giga': r'\mathrm{G}',
    'tera': r'\mathrm{T}',
    'peta': r'\mathrm{P}',
    'exa': r'\mathrm{E}',
    'zetta': r'\mathrm{Z}',
    'yotta': r'\mathrm{Y}',
    'deci': r'\mathrm{d}',
    'centi': r'\mathrm{c}',
    'milli': r'\mathrm{m}',
    'micro': r'\micro', 
    'nano': r'\mathrm{n}',
    'pico': r'\mathrm{p}',
    'femto': r'\mathrm{f}',
    'atto': r'\mathrm{a}',
    'zepto': r'\mathrm{z}',
    'yocto': r'\mathrm{y}'
}

# mapping D-SI units to latex
_dsiUnitsLatex = {
    'metre': r'\mathrm{m}',
    'kilogram': r'\mathrm{kg}',
    'second': r'\mathrm{s}',
    'ampere': r'\mathrm{A}',
    'kelvin': r'\mathrm{K}',
    'mole': r'\mathrm{mol}',
    'candela': r'\mathrm{cd}',
    'one': r'1',
    'day': r'\mathrm{d}',
    'hour': r'\mathrm{h}',
    'minute': r'\mathrm{min}',
    'degree': r'^\circ',
    'arcminute': r"'",
    'arcsecond': r"''",
    'gram': r'\mathrm{g}',
    'radian': r'\mathrm{rad}',
    'steradian': r'\mathrm{sr}',
    'hertz': r'\mathrm{Hz}',
    'newton': r'\mathrm{N}',
    'pascal': r'\mathrm{Pa}',
    'joule': r'\mathrm{J}',
    'watt': r'\mathrm{W}',
    'coulomb': r'\mathrm{C}',
    'volt': r'\mathrm{V}',
    'farad': r'\mathrm{F}',
    'ohm': r'\Omega',
    'siemens': r'\mathrm{S}',
    'weber': r'\mathrm{Wb}',
    'tesla': r'\mathrm{T}',
    'henry': r'\mathrm{H}',
    'degreecelsius': r'^\circ\mathrm{C}',
    'lumen': r'\mathrm{lm}',
    'lux': r'\mathrm{lx}',
    'becquerel': r'\mathrm{Bq}',
    'sievert': r'\mathrm{Sv}',
    'gray': r'\mathrm{Gy}',
    'katal': r'\mathrm{kat}',
    'hectare': r'\mathrm{ha}',
    'litre': r'\mathrm{l}',
    'tonne': r'\mathrm{t}',
    'electronvolt': r'\mathrm{eV}',
    'dalton': r'\mathrm{Da}',
    'astronomicalunit': r'\mathrm{au}',
    'neper': r'\mathrm{Np}',
    'bel': r'\mathrm{B}',
    'decibel': r'\mathrm{dB}'
}
    