import pytest
from dsiParser import dsiParser, _node

p = dsiParser()

def test_baseCase():
    # Most basic case: one unit without prefix or exponent
    tree = p.parse(r'\metre')
    assert tree.tree == [[_node('','metre','')]]
    assert tree.toLatex() == r'$$\mathrm{m}$$'
    assert tree.valid
    assert tree.warnings == []

    # One longer unit
    tree = p.parse(r'\milli\metre\tothe{0.5}\kilogram\per\mega\second\tothe{3}\ampere\tothe{-2}')
    assert tree.toLatex() == r'$$\frac{\sqrt{\mathrm{m}\mathrm{m}}\,\mathrm{kg}}{\mathrm{M}\mathrm{s}^{3}\,\mathrm{A}^{-2}}$$'
    assert tree.valid
    assert tree.warnings == []


def test_robustness():
    # Unknown unit
    with pytest.warns(RuntimeWarning, match='The identifier \"foo\" does not match any D-SI units!'):
        tree = p.parse(r'\foo')
        assert tree.toLatex() == r'$${\color{red}\mathrm{foo}}$$'  
        assert not tree.valid
        assert len(tree.warnings) == 1
        assert tree.warnings == ['The identifier \"foo\" does not match any D-SI units!']
    
    # Unknown string in the middle of input
    with pytest.warns(RuntimeWarning, match='The identifier \"mini\" does not match any D-SI units!'):
        tree = p.parse(r'\kilo\metre\per\mini\second')
        print(tree.toLatex())
        assert tree.toLatex() == r'$$\frac{\mathrm{k}\mathrm{m}}{{\color{red}\mathrm{mini}}\,\mathrm{s}}$$'  
        assert not tree.valid
        assert len(tree.warnings) == 1
        assert tree.warnings == ['The identifier \"mini\" does not match any D-SI units!']
    
    # Base unit missing
    with pytest.warns(RuntimeWarning, match=r'This D-SI unit seems to be missing the base unit! \"\\milli\\tothe\{2\}\"'):
        tree = p.parse(r'\milli\tothe{2}')
        print(tree.toLatex())
        assert tree.toLatex() == r'$${\color{red}\mathrm{m}{\color{red}\mathrm{}}^{2}}$$'  
        assert not tree.valid
        assert len(tree.warnings) == 1
        assert tree.warnings == ['This D-SI unit seems to be missing the base unit! \"\\milli\\tothe{2}\"']


def test_power():
    # power
    powerTree = p.parse(r'\metre\tothe{2}')
    assert powerTree.tree == [[_node('','metre','2')]]
    assert powerTree.toLatex() == r'$$\mathrm{m}^{2}$$'
    assert powerTree.valid
    assert powerTree.warnings == []
    assert p.parse(r'\metre\tothe{0.5}').toLatex() == r'$$\sqrt{\mathrm{m}}$$'
    assert p.parse(r'\metre\tothe{-2}').toLatex() == r'$$\mathrm{m}^{-2}$$'
    assert p.parse(r'\metre\tothe{1337}').toLatex() == r'$$\mathrm{m}^{1337}$$'

    # Non-numerical power
    with pytest.warns(RuntimeWarning, match='The exponent \"foo\" is not a number!'):
        assert p.parse(r'\metre\tothe{foo}').toLatex() == r'$$\mathrm{m}^{{\color{red}\mathrm{foo}}}$$'

def test_prefix():
    # D-SI prefix
    prefixTree = p.parse(r'\kilo\metre')
    assert prefixTree.tree == [[_node('kilo','metre','')]]
    assert prefixTree.toLatex() == r'$$\mathrm{k}\mathrm{m}$$'
    assert prefixTree.valid

def test_node():
    # full node
    fullNodeTree = p.parse(r'\kilo\metre\tothe{2}')
    assert fullNodeTree.tree == [[_node('kilo','metre','2')]]
    assert fullNodeTree.toLatex() == r'$$\mathrm{k}\mathrm{m}^{2}$$'
    assert fullNodeTree.valid

def test_fraction():
    fractionTree = p.parse(r'\mega\metre\per\second\tothe{2}')
    assert fractionTree.tree == [[_node('mega','metre','')],[_node('','second','2')]]
    assert fractionTree.toLatex() == r'$$\frac{\mathrm{M}\mathrm{m}}{\mathrm{s}^{2}}$$'
    assert fractionTree.valid

    # double fraction
    with pytest.warns(RuntimeWarning, match=r'The dsi string contains more than one \\per, does not match specs! Given string: \\metre\\per\\metre\\per\\metre'):
        tree = p.parse(r'\metre\per\metre\per\metre')
        assert tree.toLatex() == r'$$\mathrm{m}{\color{red}/}\mathrm{m}{\color{red}/}\mathrm{m}$$'  
        assert not tree.valid
        assert len(tree.warnings) == 1
        assert tree.warnings == [r'The dsi string contains more than one \per, does not match specs! Given string: \metre\per\metre\per\metre']

def test_empty():
    with pytest.warns(RuntimeWarning, match='Given D-SI string is empty!'):
        assert p.parse('').toLatex() == ''

def test_wrappers():
    assert p.parse(r'\metre').toLatex(wrapper='') == r'\mathrm{m}'
    assert p.parse(r'\metre').toLatex(wrapper='$', prefix=r'\mathrm{Unit: }', suffix=r'\mathrm{(generated from D-SI string)}') == r'$\mathrm{Unit: }\mathrm{m}\mathrm{(generated from D-SI string)}$'
    parserWrapper = dsiParser(latexDefaultWrapper="&")
    assert parserWrapper.parse(r'\metre').toLatex() == r'&\mathrm{m}&'
    parserFull = dsiParser(latexDefaultWrapper='@', latexDefaultPrefix=r'\mathrm{Prefix}', latexDefaultSuffix=r'\mathrm{Suffix}')
    assert parserFull.parse(r'\metre').toLatex() == r'@\mathrm{Prefix}\mathrm{m}\mathrm{Suffix}@'
